# -*- coding: utf-8 -*-

from bson import DBRef
from mongoengine.dereference import DeReference
from mongoengine.base import get_document, TopLevelDocumentMetaclass
from mongoengine.fields import (ListField, DictField, MapField,
                                ReferenceField)
from mongoengine.connection import get_db
from mongomotor.queryset import QuerySet
from mongomotor.document import Document, EmbeddedDocument
import tornado


class DeReferenceMotor(DeReference):

    @tornado.gen.coroutine
    def __call__(self, items, max_depth=1, instance=None, name=None):
        """
        Cheaply dereferences the items to a set depth.
        Also handles the conversion of complex data types.

        :param items: The iterable (dict, list, queryset) to be dereferenced.
        :param max_depth: The maximum depth to recurse to
        :param instance: The owning instance used for tracking changes by
            :class:`~mongoengine.base.ComplexBaseField`
        :param name: The name of the field, used for tracking changes by
            :class:`~mongoengine.base.ComplexBaseField`
        :param get: A boolean determining if being called by __get__
        """
        if items is None or isinstance(items, str):
            return items

        # cheapest way to convert a queryset to a list
        # list(queryset) uses a count() query to determine length
        if isinstance(items, QuerySet):
            items = [i for i in items]

        self.max_depth = max_depth
        doc_type = None

        if instance and isinstance(instance, (Document, EmbeddedDocument,
                                              TopLevelDocumentMetaclass)):
            doc_type = instance._fields.get(name)
            while hasattr(doc_type, 'field'):
                doc_type = doc_type.field

            if isinstance(doc_type, ReferenceField):
                field = doc_type
                doc_type = doc_type.document_type
                is_list = not hasattr(items, 'items')

                if is_list and all([i.__class__ == doc_type for i in items]):
                    return items
                elif not is_list and all(
                        [i.__class__ == doc_type for i in list(items.values())]):
                    return items
                elif not field.dbref:
                    if not hasattr(items, 'items'):

                        def _get_items(items):
                            new_items = []
                            for v in items:
                                if isinstance(v, list):
                                    new_items.append(_get_items(v))
                                elif not isinstance(v, (DBRef, Document)):
                                    new_items.append(field.to_python(v))
                                else:
                                    new_items.append(v)
                            return new_items

                        items = _get_items(items)
                    else:
                        items = dict([
                            (k, field.to_python(v))
                            if not isinstance(v, (DBRef, Document)) else (k, v)
                            for k, v in items.items()]
                        )

        self.reference_map = self._find_references(items)
        self.object_map = yield self._fetch_objects(doc_type=doc_type)
        return self._attach_objects(items, 0, instance, name)

    @tornado.gen.coroutine
    def _fetch_objects(self, doc_type=None):
        """Fetch all references and convert to their document objects
        """
        object_map = {}
        for collection, dbrefs in self.reference_map.items():
            keys = list(object_map.keys())
            refs = list(
                set([dbref for dbref in dbrefs if str(dbref).encode('utf-8')
                     not in keys]))
            if hasattr(collection, 'objects'):
                # We have a document class for the refs
                references = yield collection.objects.in_bulk(refs)
                for key, doc in references.items():
                    object_map[key] = doc
            else:
                # Generic reference: use the refs data to convert to document
                if isinstance(doc_type, (ListField, DictField, MapField)):
                    continue

                if doc_type:
                    references = doc_type._get_db()[collection].find(
                        {'_id': {'$in': refs}})

                    next_ref = yield self._get_next_doc(references)
                    while next_ref:
                        doc = doc_type._from_son(next_ref)
                        object_map[(collection, doc.id)] = doc
                        next_ref = yield self._get_next_doc(references)

                else:
                    references = get_db()[collection].find(
                        {'_id': {'$in': refs}})
                    ref = yield self._get_next_doc(references)
                    while ref:
                        if '_cls' in ref:
                            doc = get_document(ref["_cls"])._from_son(ref)
                        elif doc_type is None:
                            doc = get_document(
                                ''.join(x.capitalize()
                                        for x in collection.split('_')))
                            doc = doc._from_son(ref)
                        else:
                            doc = doc_type._from_son(ref)
                        object_map[(collection, doc.id)] = doc
                        ref = yield self._get_next_doc(references)

        return object_map

    @tornado.gen.coroutine
    def _get_next_doc(self, cursor):
        yield cursor.fetch_next
        n = cursor.next_object()
        return n



# -*- coding: utf-8 -*-

import collections
from mongoengine.base.metaclasses import TopLevelDocumentMetaclass, MetaDict
from mongoengine.errors import DoesNotExist, MultipleObjectsReturned
from mongoengine.fields import DynamicField
from mongomotor.queryset import QuerySetManager


class MapReduceDocumentMetaclass(TopLevelDocumentMetaclass):

    """ Metaclass for map-reduced collections.
    """

    def __new__(cls, name, bases, attrs):
        flattened_bases = cls._get_bases(bases)
        super_new = super(TopLevelDocumentMetaclass, cls).__new__

        # Set default _meta data if base class, otherwise get user defined meta

        if (attrs.get('my_metaclass') == MapReduceDocumentMetaclass):
            # defaults
            attrs['_meta'] = {
                'abstract': True,
                'max_documents': None,
                'max_size': None,
                'ordering': [],  # default ordering applied at runtime
                'indexes': [],  # indexes to be ensured at runtime
                'id_field': None,
                'index_background': False,
                'index_drop_dups': False,
                'index_opts': None,
                'delete_rules': None,
                'allow_inheritance': None,
            }
            attrs['_is_base_cls'] = True
            attrs['_meta'].update(attrs.get('meta', {}))
        else:
            attrs['_meta'] = attrs.get('meta', {})
            # Explictly set abstract to false unless set
            attrs['_meta']['abstract'] = attrs['_meta'].get('abstract', False)
            attrs['_is_base_cls'] = False

        # Set flag marking as document class - as opposed to an object mixin
        attrs['_is_document'] = True

        # Find the parent document class
        parent_doc_cls = [b for b in flattened_bases
                          if b.__class__ == MapReduceDocumentMetaclass]
        parent_doc_cls = None if not parent_doc_cls else parent_doc_cls[0]

        # Ensure abstract documents have abstract bases
        if attrs.get('_is_base_cls') or attrs['_meta'].get('abstract'):
            return super_new(cls, name, bases, attrs)

        # Merge base class metas.
        # Uses a special MetaDict that handles various merging rules
        meta = MetaDict()
        for base in flattened_bases[::-1]:
            # Add any mixin metadata from plain objects
            if hasattr(base, '_meta'):
                meta.merge(base._meta)

        meta.merge(attrs.get('_meta', {}))  # Top level meta

        # Set default collection name
        if 'collection' not in meta:
            meta['collection'] = ''.join('_%s' % c if c.isupper() else c
                                         for c in name).strip('_').lower()
        attrs['_meta'] = meta

        # Call super and get the new class
        new_class = super_new(cls, name, bases, attrs)

        meta = new_class._meta

        # Set index specifications
        meta['index_specs'] = new_class._build_index_specs(meta['indexes'])

        # Provide a default queryset unless exists or one has been set
        if 'objects' not in dir(new_class):
            new_class.objects = QuerySetManager()

        # Set primary key if not defined by the document
        new_class._auto_id_field = getattr(parent_doc_cls,
                                           '_auto_id_field', False)
        if not new_class._meta.get('id_field'):
            new_class._auto_id_field = True
            new_class._meta['id_field'] = 'id'
            new_class._fields['id'] = DynamicField(db_field='_id')
            new_class._fields['id'].name = 'id'
            new_class.id = new_class._fields['id']

        # Prepend id field to _fields_ordered
        if 'id' in new_class._fields and 'id' not in new_class._fields_ordered:
            new_class._fields_ordered = ('id', ) + new_class._fields_ordered

        # Merge in exceptions with parent hierarchy
        exceptions_to_merge = (DoesNotExist, MultipleObjectsReturned)
        module = attrs.get('__module__')
        for exc in exceptions_to_merge:
            name = exc.__name__
            parents = tuple(getattr(base, name) for base in flattened_bases
                            if hasattr(base, name)) or (exc,)
            # Create new exception and set to new_class
            exception = type(name, parents, {'__module__': module})
            setattr(new_class, name, exception)

        return new_class



# -*- coding: utf-8 -*-

from mongomotor.asyncsignals import Namespace

signals_available = True

_signals = Namespace()

pre_init = _signals.signal('pre_init')
post_init = _signals.signal('post_init')
pre_save = _signals.signal('pre_save')
pre_save_post_validation = _signals.signal('pre_save_post_validation')
post_save = _signals.signal('post_save')
pre_delete = _signals.signal('pre_delete')
post_delete = _signals.signal('post_delete')
pre_bulk_insert = _signals.signal('pre_bulk_insert')
post_bulk_insert = _signals.signal('post_bulk_insert')



# -*- coding: utf-8 -*-

import tornado
from tornado import gen
from mongoengine.queryset import visitor as visitor_module
from mongoengine.queryset import transform


class QueryCompilerVisitor(visitor_module.QueryCompilerVisitor):

    @gen.coroutine
    def visit_query(self, query):
        r = yield transform.query(self.document, **query.query)
        return r


class QNodeBase:
    @gen.coroutine
    def to_query(self, document):
        query = yield self.accept(visitor_module.SimplificationVisitor())
        query = yield query.accept(QueryCompilerVisitor(document))
        return query


class Q(QNodeBase, visitor_module.Q):

    @gen.coroutine
    def accept(self, visitor):
        r = yield visitor.visit_query(self)
        return r


class QCombination(QNodeBase, visitor_module.QCombination):

    @gen.coroutine
    def accept(self, visitor):
        for i in range(len(self.children)):
            if isinstance(self.children[i], visitor_module.QNode):
                accepted = self.children[i].accept(visitor)
                if isinstance(accepted, tornado.concurrent.Future):
                    accepted = yield accepted
                self.children[i] = accepted

        return visitor.visit_combination(self)



# -*- coding: utf-8 -*-

from bson import DBRef
from tornado import gen
from tornado.concurrent import Future
from mongoengine.common import _import_class
from mongoengine import fields
from mongoengine.base.datastructures import (
    BaseDict, BaseList, EmbeddedDocumentList
)
from mongoengine.fields import *
from mongoengine import Document


class ComplexBaseField(fields.ComplexBaseField):

    def __get__(self, instance, owner):
        """Descriptor to automatically dereference references.
        """
        if instance is None:
            # Document class being used rather than a document object
            return self

        ReferenceField = _import_class('ReferenceField')
        GenericReferenceField = _import_class('GenericReferenceField')
        dereference = (self._auto_dereference and
                       (self.field is None or isinstance(
                           self.field,
                           (GenericReferenceField, ReferenceField))))

        _dereference = _import_class("DeReference")()

        self._auto_dereference = instance._fields[self.name]._auto_dereference
        initialised = instance._initialised
        is_refcls = instance._data.get(self.name) and not isinstance(
            instance._data.get(self.name), Future)
        is_dbref = instance._data.get(self.name) and bool(
            [v for v in instance._data.get(self.name) if isinstance(v, DBRef)])

        if is_dbref or (initialised and dereference and
                        instance._data.get(self.name) and not is_refcls):
            @gen.coroutine
            def deref(instance):
                instance._data[self.name] = yield _dereference(
                    instance._data.get(self.name), max_depth=1,
                    instance=instance, name=self.name)

                value = super(fields.ComplexBaseField, self).__get__(
                    instance, owner)
                return self._convert_collections(instance, value)

            return deref(instance)
        else:
            value = super(fields.ComplexBaseField, self).__get__(
                instance, owner)
            return self._convert_collections(instance, value)

    def _convert_collections(self, instance, value):

        # Convert lists / values so we can watch for any changes on them
        EmbeddedDocumentListField = _import_class('EmbeddedDocumentListField')
        if isinstance(value, (list, tuple)):
            if (issubclass(type(self), EmbeddedDocumentListField) and
                    not isinstance(value, EmbeddedDocumentList)):
                value = EmbeddedDocumentList(value, instance, self.name)
            elif not isinstance(value, BaseList):
                value = BaseList(value, instance, self.name)
            instance._data[self.name] = value
        elif isinstance(value, dict) and not isinstance(value, BaseDict):
            value = BaseDict(value, instance, self.name)
            instance._data[self.name] = value

        is_refcls = instance._data.get(self.name) and not isinstance(
            instance._data.get(self.name), Future)

        _dereference = _import_class("DeReference")()
        if (self._auto_dereference and instance._initialised and
                isinstance(value, (BaseList, BaseDict)) and
                not value._dereferenced and not is_refcls):

            @gen.coroutine
            def deref(instance, value):
                value = yield _dereference(
                    value, max_depth=1, instance=instance, name=self.name
                )
                value._dereferenced = True
                instance._data[self.name] = value
                return value
            return deref(instance, value)

        else:
            return value


class ReferenceField(fields.ReferenceField):

    def __get__(self, instance, owner):
        """Descriptor to allow lazy dereferencing.
        """
        if instance is None:
            # Document class being used rather than a document object
            return self

        # Get value from document instance if available
        value = instance._data.get(self.name)
        # self._auto_dereference = True
        self._auto_dereference = instance._fields[self.name]._auto_dereference

        # Dereference DBRefs
        if self._auto_dereference and isinstance(value, DBRef):
            @gen.coroutine
            def deref(value):

                db = self.document_type._get_db()

                value = yield db.dereference(value)
                if value is not None:
                    instance._data[self.name] = self.document_type._from_son(
                        value)
                return super(fields.ReferenceField, self).__get__(
                    instance, owner)
            return deref(value)
        else:
            return super(fields.ReferenceField, self).__get__(instance, owner)


class ListField(ComplexBaseField, fields.ListField):
    pass


class DictField(ComplexBaseField, fields.DictField):
    pass



# -*- coding: utf-8 -*-

import re
import pymongo
from pymongo.read_preferences import ReadPreference
from bson.dbref import DBRef
import tornado
from tornado import gen
from mongoengine.queryset import OperationError, NotUniqueError
from mongoengine import (Document as DocumentBase,
                         EmbeddedDocument as EmbeddedDocumentBase,
                         DynamicDocument as DynamicDocumentBase)
from mongoengine.base.metaclasses import (TopLevelDocumentMetaclass,
                                          DocumentMetaclass)
from mongoengine.document import _import_class, includes_cls
from mongomotor import signals
from mongomotor.base.metaclasses import MapReduceDocumentMetaclass


class Document(DocumentBase, metaclass=TopLevelDocumentMetaclass):
    """
    Document version that uses motor mongodb driver.
    It's a copy of some mongoengine.Document methods
    that uses tornado.gen.coroutine and yield things
    to use motor with generator style.
    """

    my_metaclass = TopLevelDocumentMetaclass

    @gen.coroutine
    def delete(self, **write_concern):
        """Delete the :class:`~mongoengine.Document` from the database. This
        will only take effect if the document has been previously saved.

        :param write_concern: Extra keyword arguments are passed down which
            will be used as options for the resultant
            ``getLastError`` command.  For example,
            ``save(..., write_concern={w: 2, fsync: True}, ...)`` will
            wait until at least two servers have recorded the write and
            will force an fsync on the primary server.
        """
        signals.pre_delete.send(self.__class__, document=self)
        try:
            qs = self._qs.filter(**self._object_key)
            yield qs.delete(write_concern=write_concern, _from_doc_delete=True)
        except pymongo.errors.OperationFailure as err:
            message = 'Could not delete document (%s)' % err.message
            raise OperationError(message)
        signals.post_delete.send(self.__class__, document=self)

    @gen.coroutine
    def save(self, force_insert=False, validate=True, clean=True,
             write_concern=None,  cascade=None, cascade_kwargs=None,
             _refs=None, **kwargs):
        """Save the :class:`~mongoengine.Document` to the database. If the
        document already exists, it will be updated, otherwise it will be
        created.

        :param force_insert: only try to create a new document, don't allow
            updates of existing documents
        :param validate: validates the document; set to ``False`` to skip.
        :param clean: call the document clean method, requires `validate` to be
            True.
        :param write_concern: Extra keyword arguments are passed down to
            :meth:`~pymongo.collection.Collection.save` OR
            :meth:`~pymongo.collection.Collection.insert`
            which will be used as options for the resultant
            ``getLastError`` command.  For example,
            ``save(..., write_concern={w: 2, fsync: True}, ...)`` will
            wait until at least two servers have recorded the write and
            will force an fsync on the primary server.
        :param cascade: Sets the flag for cascading saves.  You can set a
            default by setting "cascade" in the document __meta__
        :param cascade_kwargs: (optional) kwargs dictionary to be passed throw
            to cascading saves.  Implies ``cascade=True``.
        :param _refs: A list of processed references used in cascading saves

        """
        signals.pre_save.send(self.__class__, document=self)

        if validate:
            self.validate(clean=clean)

        if write_concern is None:
            write_concern = {"w": 1}

        doc = self.to_mongo()

        created = ('_id' not in doc or self._created or force_insert)

        signals.pre_save_post_validation.send(self.__class__, document=self,
                                              created=created)

        try:
            collection = self._get_collection()
            if created:
                if force_insert:
                    # I realy don't know why this stupid line isn't covered. Maybe later
                    # I take a look at it.
                    object_id = yield collection.insert(doc, **write_concern)  # pragma: no cover
                else:
                    object_id = yield collection.save(doc, **write_concern)
            else:
                object_id = doc['_id']
                updates, removals = self._delta()
                # Need to add shard key to query, or you get an error
                select_dict = {'_id': object_id}
                shard_key = self.__class__._meta.get('shard_key', tuple())
                for k in shard_key:
                    actual_key = self._db_field_map.get(k, k)
                    select_dict[actual_key] = doc[actual_key]

                def is_new_object(last_error):
                    if last_error is not None:
                        updated = last_error.get("updatedExisting")
                        if updated is not None:
                            return not updated
                    return created  # pragma: no cover

                update_query = {}

                if updates:
                    update_query["$set"] = updates
                if removals:
                    update_query["$unset"] = removals  # pragma: no cover
                if updates or removals:
                    last_error = yield collection.update(select_dict,
                                                         update_query,
                                                         upsert=True,
                                                         **write_concern)
                    created = is_new_object(last_error)

            if cascade is None:
                cascade = self._meta.get('cascade', False) or \
                    cascade_kwargs is not None

            if cascade:
                kwargs = {
                    "force_insert": force_insert,
                    "validate": validate,
                    "write_concern": write_concern,
                    "cascade": cascade
                }
                if cascade_kwargs:  # Allow granular control over cascades
                    kwargs.update(cascade_kwargs)
                kwargs['_refs'] = _refs
                yield self.cascade_save(**kwargs)

        except pymongo.errors.OperationFailure as err:
            message = 'Could not save document (%s)'
            if re.match('^E1100[01] duplicate key', str(err)):
                # E11000 - duplicate key error index
                # E11001 - duplicate key on update
                message = 'Tried to save duplicate unique keys (%s)'
                raise NotUniqueError(message % str(err))
            raise OperationError(message % str(err))  # pragma: no cover
        id_field = self._meta['id_field']
        if id_field not in self._meta.get('shard_key', []):
            self[id_field] = self._fields[id_field].to_python(object_id)

        self._clear_changed_fields()
        self._created = False
        signals.post_save.send(self.__class__, document=self, created=created)

        return self

    @gen.coroutine
    def cascade_save(self, *args, **kwargs):
        """Recursively saves any references /
           generic references on an objects"""
        _refs = kwargs.get('_refs', []) or []

        ReferenceField = _import_class('ReferenceField')
        GenericReferenceField = _import_class('GenericReferenceField')

        for name, cls in list(self._fields.items()):
            if not isinstance(cls, (ReferenceField,
                                    GenericReferenceField)):
                continue

            ref = self._data.get(name)
            if not ref or isinstance(ref, DBRef):
                continue  # pragma: no cover

            if not getattr(ref, '_changed_fields', True):
                continue  # pragma: no cover

            ref_id = "%s,%s" % (ref.__class__.__name__, str(ref._data))
            if ref and ref_id not in _refs:
                _refs.append(ref_id)
                kwargs["_refs"] = _refs
                yield ref.save(**kwargs)
                ref._changed_fields = []

    @classmethod
    @gen.coroutine
    def drop_collection(cls):
        """Drops the entire collection associated with this
        :class:`~mongoengine.Document` type from the database.
        """
        cls._collection = None
        db = cls._get_db()
        yield db.drop_collection(cls._get_collection_name())

    @classmethod
    @gen.coroutine
    def compare_indexes(cls):
        """ Compares the indexes defined in MongoEngine with the ones existing
        in the database. Returns any missing/extra indexes.
        """
        required = cls.list_indexes()
        existing = [
            info['key'] for info in
            list((yield cls._get_collection().index_information()).values())]

        missing = [index for index in required if index not in existing]
        extra = [index for index in existing if index not in required]

        # if { _cls: 1 } is missing, make sure it's *really* necessary
        if [('_cls', 1)] in missing:
            cls_obsolete = False
            for index in existing:
                if includes_cls(index) and index not in extra:
                    cls_obsolete = True
                    break
            if cls_obsolete:
                missing.remove([('_cls', 1)])

        return {'missing': missing, 'extra': extra}

    @gen.coroutine
    def reload(self, max_depth=1):
        """Reloads all attributes from the database.

        .. versionadded:: 0.1.2
        .. versionchanged:: 0.6  Now chainable
        """
        if not self.pk:
            raise self.DoesNotExist("Document does not exist")
        obj = (yield (yield self._qs.read_preference(ReadPreference.PRIMARY).filter(
            **self._object_key).limit(1)).select_related(max_depth=max_depth))

        if obj:
            obj = obj[0]
        else:
            raise self.DoesNotExist("Document does not exist")

        for field in self._fields_ordered:
            if isinstance(obj[field], tornado.concurrent.Future):

                # this will mark document as changed in this field but this is not
                # the desired behavior, so we'll remove this mark after.
                f = yield self._load_related(obj[field], max_depth - 1)
                obj[field] = f
                # obj._changed_fields.pop(obj._changed_fields.index(field))
                setattr(self, field, self._reload(field, f))
            else:
                setattr(self, field, self._reload(field, obj[field]))

        self._changed_fields = obj._changed_fields
        self._created = False
        return obj

    @gen.coroutine
    def _load_related(self, field, max_depth):
        field = yield field
        return field
        # if max_depth == 0:
        #     return field
        # else:
        #     for f in dir(field):


class DynamicDocument(Document, DynamicDocumentBase,
                      metaclass=TopLevelDocumentMetaclass):

    my_metaclass = TopLevelDocumentMetaclass

    _dynamic = True

    def __delattr__(self, *args, **kwargs):
        DynamicDocumentBase.__delattr__(self, *args, **kwargs)


class EmbeddedDocument(EmbeddedDocumentBase,
                       metaclass=DocumentMetaclass):

    my_metaclass = TopLevelDocumentMetaclass

    def __init__(self, *args, **kwargs):
        super(EmbeddedDocument, self).__init__(*args, **kwargs)
        self._instance = None
        self._changed_fields = []


class MapReduceDocument(DynamicDocument, metaclass=MapReduceDocumentMetaclass):
    """This MapReduceDocument is different from the mongoengine's one
    because its intent is to allow you to query over."""

    my_metaclass = MapReduceDocumentMetaclass
